# coding: utf-8

from urllib.parse import urljoin

from geoserver.support import ResourceInfo, xml_property, write_bool
from geoserver.style import Style
from geoserver import settings


class _Attribution:
    def __init__(self, title, width, height, href, url, logo_type):
        self.title = title
        self.width = width
        self.height = height
        self.href = href
        self.url = url
        self.logo_type = logo_type

def _read_attribution(node):
    title = node.find("title")
    width = node.find("logoWidth")
    height = node.find("logoHeight")
    href = node.find("href")
    url = node.find("logoURL")
    logo_type = node.find("logoType")
    if title is not None:
        title = title.text
    if width is not None:
        width = width.text
    if height is not None:
        height = height.text
    if href is not None:
        href = href.text
    if url is not None:
        url = url.text
    if logo_type is not None:
        logo_type = logo_type.text
    return _Attribution(title, width, height, href, url, logo_type)


def _write_attribution(builder, attribution):
    builder.start("attribution", dict())
    if attribution.title is not None:
        builder.start("title", dict())
        builder.data(attribution.title)
        builder.end("title")
    if attribution.width is not None:
        builder.start("logoWidth", dict())
        builder.data(attribution.width)
        builder.end("logoWidth")
    if attribution.height is not None:
        builder.start("logoHeight", dict())
        builder.data(attribution.height)
        builder.end("logoHeight")
+    if attr.href is not None:
        builder.start("href", dict())
        builder.data(attr.href)
        builder.end("href")
    if attr.url is not None:
        builder.start("logoURL", dict())
        builder.data(attr.url)
        builder.end("logoURL")
    if attr.type is not None:
        builder.start("logoType", dict())
        builder.data(attr.type)
        builder.end("logoType")
    builder.end("attribution")


def _write_style_element(builder, name):
    ws, name = name.split(':') if ':' in name else (None, name)
    builder.start("name", dict())
    builder.data(name)
    builder.end("name")
    if ws:
        builder.start("workspace", dict())
        builder.data(ws)
        builder.end("workspace")


def _write_default_style(builder, name):
    builder.start("defaultStyle", dict())
    if name is not None:
        _write_style_element(builder, name)
    builder.end("defaultStyle")


def _write_alternate_styles(builder, styles):
    builder.start("styles", dict())
    for s in styles:
        builder.start("style", dict())
        _write_style_element(builder, getattr(s, 'fqn', s))
        builder.end("style")
    builder.end("styles")


class Layer(ResourceInfo):
    def __init__(self, catalog, name):
        super(Layer, self).__init__()
        self.catalog = catalog
        self.name = name

    resource_type = "layer"
    save_method = settings.PUT

    @property
    def href(self):
        return urljoin(
            self.catalog.service_url,
            "layers/{}.xml".format(self.name)
        )

    @property
    def resource(self):
        if self.dom is None:
            self.fetch()
        name = self.dom.find("resource/name").text
        return self.catalog.get_resource(name)

    @property
    def default_style(self):
        if 'default_style' in self.dirty:
            return self.dirty['default_style']
        if self.dom is None:
            self.fetch()
        element = self.dom.find("defaultStyle")
        # aborted data uploads can result in no default style
        return self._resolve_style(element) if element is not None else None

    @default_style.setter
    def default_style(self, style):
        if isinstance(style, Style):
            style = style.fqn
        self.dirty["default_style"] = style

    @property
    def styles(self):
        if "alternate_styles" in self.dirty:
            return self.dirty["alternate_styles"]
        if self.dom is None:
            self.fetch()
        styles_list = self.dom.findall("styles/style")
        return filter(None, [self._resolve_style(s) for s in styles_list])

    @styles.setter
    def styles(self, styles):
        self.dirty["alternate_styles"] = styles

    def _resolve_style(self, element):
        # instead of using name or the workspace element (which only appears
        # in >=2.4), just use the atom link href attribute
        atom_link = [n for n in element.getchildren() if 'href' in n.attrib]
        if atom_link:
            style_workspace_url = atom_link[0].attrib.get("href")
            return self.catalog.get_style_by_url(style_workspace_url)

    @property
    def attribution(self):
        return self.attribution_object.title

    @attribution.setter
    def attribution(self, text):
        self.dirty["attribution"] = _Attribution(
            text,
            self.attribution_object.width,
            self.attribution_object.height
        )
        assert self.attribution_object.title == text

    attribution_object = xml_property("attribution", _read_attribution)
    enabled = xml_property("enabled", lambda x: x.text == "true")
    advertised = xml_property("advertised", lambda x: x.text == "true",
                              default=True)

    def _get_attr_attribution(self):
        return { 'title': self.attribution_object.title,
                 'width': self.attribution_object.width,
                 'height': self.attribution_object.height,
                 'href': self.attribution_object.href,
                 'url': self.attribution_object.url,
                 'type': self.attribution_object.logo_type }

    def _set_attr_attribution(self, attribution):
        self.dirty["attribution"] = _attribution( attribution['title'],
                                                  attribution['width'],
                                                  attribution['height'],
                                                  attribution['href'],
                                                  attribution['url'],
                                                  attribution['type'] )

        assert self.attribution_object.title == attribution['title']
        assert self.attribution_object.width == attribution['width']
        assert self.attribution_object.height == attribution['height']
        assert self.attribution_object.href == attribution['href']
        assert self.attribution_object.url == attribution['url']
        assert self.attribution_object.type == attribution['type']

    attribution = property(_get_attr_attribution, _set_attr_attribution)

    writers = {
        'attribution': _write_attribution,
        'enabled': write_bool("enabled"),
        'advertised': write_bool("advertised"),
        'default_style': _write_default_style,
        'alternate_styles': _write_alternate_styles
    }



# coding: utf-8

from urllib.parse import urljoin

from geoserver.support import ResourceInfo, xml_property


class Style(ResourceInfo):
    supported_formats = ['sld10', 'sld11', 'zip']
    content_types = {
        "sld10": "application/vnd.ogc.sld+xml",
        "sld11": "application/vnd.ogc.se+xml",
        "zip": "application/zip"
    }

    def __init__(self, catalog, name, workspace=None, style_format='sld10'):
        super(Style, self).__init__()
        assert isinstance(name, str)
        assert style_format in self.supported_formats
        self.catalog = catalog
        self.workspace = workspace
        self.name = name
        self.style_format = style_format
        self._sld_dom = None

    @property
    def fqn(self):
        if not self.workspace:
            return self.name
        return '{}:{}'.format(self.workspace, self.name)

    @property
    def href(self):
        return self._build_href('.xml')

    @property
    def body_href(self):
        return self._build_href('.sld')

    @property
    def create_href(self):
        return self._build_href('.xml', True)

    @property
    def content_type(self):
        return Style.content_types[self.style_format]

    def _build_href(self, extension, create=False):
        url_part = "styles/"
        if not create:
            url_part += "{}{}".format(self.name, extension)
        else:
            url_part += "?name={}".format(self.name)
        if self.workspace is not None:
            url_part = "workspaces/{}/{}".format(
                getattr(self.workspace, 'name', self.workspace),
                url_part
            )
        return urljoin(
            self.catalog.service_url,
            url_part
        )

    filename = xml_property("filename")

    def _get_sld_dom(self):
        if self._sld_dom is None:
            self._sld_dom = self.catalog.get_xml(self.body_href)
        return self._sld_dom

    @property
    def sld_title(self):
        user_style = self._get_sld_dom().find("{http://www.opengis.net/sld}NamedLayer/{http://www.opengis.net/sld}UserStyle")
        if not user_style:
            user_style = self._get_sld_dom().find("{http://www.opengis.net/sld}UserLayer/{http://www.opengis.net/sld}UserStyle")
        if user_style:
            try:
                # it is not mandatory
                title_node = user_style.find("{http://www.opengis.net/sld}Title")
            except:
                title_node = None
        if title_node is not None:
            title_node = title_node.text
        return title_node

    @property
    def sld_name(self):
        user_style = self._get_sld_dom().find("{http://www.opengis.net/sld}NamedLayer/{http://www.opengis.net/sld}UserStyle")
        if not user_style:
            user_style = self._get_sld_dom().find("{http://www.opengis.net/sld}UserLayer/{http://www.opengis.net/sld}UserStyle")
        if user_style:
            try:
                # it is not mandatory
                name_node = user_style.find("{http://www.opengis.net/sld}Name")
            except:
                name_node = None
        if name_node is not None:
            name_node = name_node.text
        return name_node

    @property
    def sld_body(self):
        return self.catalog.session.get(self.body_href).text

    def update_body(self, body):
        headers = {"Content-Type": self.content_type}
        self.catalog.session.put(self.body_href, data=body, headers=headers)



# coding: utf-8

"""
gsconfig is a python library for manipulating a GeoServer instance via the GeoServer RESTConfig API.

The project is distributed under a MIT License .
"""

from urllib.parse import urljoin

from geoserver import settings
from geoserver.support import ResourceInfo, xml_property, write_string, \
    bbox, metadata, write_metadata, write_bbox, string_list, \
    write_string_list, attribute_list, write_bool


def md_link(node):
    """Extract a metadata link tuple from an xml node"""
    mimetype = node.find("type")
    mdtype = node.find("metadataType")
    content = node.find("content")
    if None in [mimetype, mdtype, content]:
        return None
    else:
        return mimetype.text, mdtype.text, content.text


def metadata_link_list(node):
    if node is not None:
        return [md_link(n) for n in node.findall("metadataLink")]


def write_metadata_link_list(name):
    def write(builder, md_links):
        builder.start(name, dict())
        if md_links:
            for (mime, md_type, content_url) in md_links:
                # geoserver supports only three mime
                if md_type not in ['ISO19115:2003', 'FGDC', 'TC211']:
                    mime = 'other'
                    md_type = 'other'
                builder.start("metadataLink", dict())
                builder.start("type", dict())
                builder.data(mime)
                builder.end("type")
                builder.start("metadataType", dict())
                builder.data(md_type)
                builder.end("metadataType")
                builder.start("content", dict())
                builder.data(content_url)
                builder.end("content")
                builder.end("metadataLink")
        builder.end(name)
    return write


def featuretype_from_index(catalog, workspace, store, node):
    name = node.find("name")
    return FeatureType(catalog, workspace, store, name.text)


def coverage_from_index(catalog, workspace, store, node):
    name = node.find("name")
    return Coverage(catalog, workspace, store, name.text)


def wmslayer_from_index(catalog, workspace, store, node):
    name = node.find("name")
    return WmsLayer(catalog, workspace, store, name.text)


class _ResourceBase(ResourceInfo):
    save_method = settings.PUT

    def __init__(self, catalog, workspace, store, name, href=None):
        super(_ResourceBase, self).__init__()
        if not href:
            assert isinstance(store, ResourceInfo)
            assert isinstance(name, str)
            assert workspace is not None
        else:
            parts = href.split('/')
            self._workspace_name = parts[parts.index('workspaces') + 1]
            self._store_name = parts[parts.index(self.url_part_stores) + 1]
            name = parts[-1].replace('.xml','')
        self._href = href
        self.catalog = catalog
        self._workspace = workspace
        self._store = store
        self.name = name

    @property
    def workspace(self):
        if not self._workspace:
            self._workspace = self.catalog.get_workspace(self._workspace_name)
        return self._workspace

    @property
    def store(self):
        if not self._store:
            self._store = self.catalog.get_store(self._store_name, self._workspace_name)
        return self._store

    @property
    def href(self):
        if self._href:
            return self._href
        return urljoin(
            self.catalog.service_url,
            "workspaces/{}/{}/{}/{}/{}.xml".format(
                self.workspace.name,
                self.url_part_stores,
                self.store.name,
                self.url_part_types,
                self.name
            )
        )


class FeatureType(_ResourceBase):

    resource_type = "featureType"
    url_part_stores = 'datastores'
    url_part_types = 'featuretypes'

    title = xml_property("title")
    native_name = xml_property("nativeName")
    abstract = xml_property("abstract")
    enabled = xml_property("enabled")
    advertised = xml_property("advertised", default="true")
    native_bbox = xml_property("nativeBoundingBox", bbox)
    latlon_bbox = xml_property("latLonBoundingBox", bbox)
    projection = xml_property("srs")
    projection_policy = xml_property("projectionPolicy")
    keywords = xml_property("keywords", string_list)
    attributes = xml_property("attributes", attribute_list)
    metadata_links = xml_property("metadataLinks", metadata_link_list)
    metadata = xml_property("metadata", metadata)

    writers = {
        'name': write_string("name"),
        'nativeName': write_string("nativeName"),
        'title': write_string("title"),
        'abstract': write_string("abstract"),
        'enabled': write_bool("enabled"),
        'advertised': write_bool("advertised"),
        'nativeBoundingBox': write_bbox("nativeBoundingBox"),
        'latLonBoundingBox': write_bbox("latLonBoundingBox"),
        'srs': write_string("srs"),
        'nativeCRS': write_string("nativeCRS"),
        'projectionPolicy': write_string("projectionPolicy"),
        'keywords': write_string_list("keywords"),
        'metadataLinks': write_metadata_link_list("metadataLinks"),
        'metadata': write_metadata("metadata")
    }


class CoverageDimension(object):
    def __init__(self, name, description, dimension_range):
        self.name = name
        self.description = description
        self.dimension_range = dimension_range


def coverage_dimension(node):
    name = node.find("name")
    if name is not None:
        name = name.text
    description = node.find("description")
    if description is not None:
        description = description.text
    range_min = node.find("range/min")
    range_max = node.find("range/max")
    dimension_range = None
    if None not in [range_min, range_min]:  # instead of [min, max] ?!
        dimension_range = float(range_min.text), float(range_max.text)
    if None not in [name, description]:
        return CoverageDimension(name, description, dimension_range)
    else:
        return None  # should we bomb out more spectacularly here?


def coverage_dimension_xml(builder, dimension):
    # coverage dimension
    builder.start("coverageDimension", dict())
    builder.start("name", dict())
    builder.data(dimension.name)
    builder.end("name")
    # description
    builder.start("description", dict())
    builder.data(dimension.description)
    builder.end("description")
    # range
    if dimension.range is not None:
        builder.start("range", dict())
        builder.start("min", dict())
        builder.data(str(dimension.range[0]))
        builder.end("min")
        builder.start("max", dict())
        builder.data(str(dimension.range[1]))
        builder.end("max")
        builder.end("range")
    builder.end("coverageDimension")


class Coverage(_ResourceBase):

    resource_type = "coverage"
    url_part_stores = 'coveragestores'
    url_part_types = 'coverages'

    title = xml_property("title")
    abstract = xml_property("abstract")
    enabled = xml_property("enabled")
    advertised = xml_property("advertised", default="true")
    native_bbox = xml_property("nativeBoundingBox", bbox)
    latlon_bbox = xml_property("latLonBoundingBox", bbox)
    projection = xml_property("srs")
    projection_policy = xml_property("projectionPolicy")
    keywords = xml_property("keywords", string_list)
    request_srs_list = xml_property("requestSRS", string_list)
    response_srs_list = xml_property("responseSRS", string_list)
    supported_formats = xml_property("supportedFormats", string_list)
    metadata_links = xml_property("metadataLinks", metadata_link_list)
    metadata = xml_property("metadata", metadata)

    writers = {
        'title': write_string("title"),
        'abstract': write_string("abstract"),
        'enabled': write_bool("enabled"),
        'advertised': write_bool("advertised"),
        'nativeBoundingBox': write_bbox("nativeBoundingBox"),
        'latLonBoundingBox': write_bbox("latLonBoundingBox"),
        'srs': write_string("srs"),
        'projection_policy': write_string("projectionPolicy"),
        'keywords': write_string_list("keywords"),
        'metadataLinks': write_metadata_link_list("metadataLinks"),
        'requestSRS': write_string_list("requestSRS"),
        'responseSRS': write_string_list("responseSRS"),
        'supportedFormats': write_string_list("supportedFormats"),
        'metadata': write_metadata("metadata")
    }


class WmsLayer(ResourceInfo):
    resource_type = "wmsLayer"
    save_method = settings.PUT

    def __init__(self, catalog, workspace, store, name):
        super(WmsLayer, self).__init__()
        self.catalog = catalog
        self.workspace = workspace
        self.store = store
        self.name = name

    @property
    def href(self):
        return urljoin(
            self.catalog.service_url,
            "workspaces/{}/wmsstores/{}/wmslayers/{}.xml".format(
                self.workspace.name,
                self.store.name,
                self.name
            )
        )

    title = xml_property("title")
    description = xml_property("description")
    abstract = xml_property("abstract")
    keywords = xml_property("keywords", string_list)
    # nativeCRS
    projection = xml_property("srs")
    native_bbox = xml_property("nativeBoundingBox", bbox)
    latlon_bbox = xml_property("latLonBoundingBox", bbox)
    projection_policy = xml_property("projectionPolicy")
    enabled = xml_property("enabled", lambda x: x.text == "true")
    advertised = xml_property("advertised", lambda x: x.text == "true",
                              default=True)
    metadata_links = xml_property("metadataLinks", metadata_link_list)

    writers = {
        'title': write_string("title"),
        'description': write_string("description"),
        'abstract': write_string("abstract"),
        'keywords': write_string_list("keywords"),
        # nativeCRS
        'srs': write_string("srs"),
        'nativeBoundingBox': write_bbox("nativeBoundingBox"),
        'latLonBoundingBox': write_bbox("latLonBoundingBox"),
        'projectionPolicy': write_string("projectionPolicy"),
        'enabled': write_bool("enabled"),
        'advertised': write_bool("advertised"),
        'metadataLinks': write_metadata_link_list("metadataLinks")
    }



# coding: utf-8

"""
gsconfig is a python library for manipulating a GeoServer instance via
the GeoServer RESTConfig API.

The project is distributed under a MIT License .
"""


import logging
from tempfile import mkstemp
from zipfile import ZipFile
import os
from xml.etree.ElementTree import TreeBuilder, tostring


LOGGER = logging.getLogger("gsconfig2.support")


FORCE_DECLARED = "FORCE_DECLARED"
# The projection handling policy for layers that should use coordinates
# directly while reporting the configured projection to clients.  This should be
# used when projection information is missing from the underlying datastore.


FORCE_NATIVE = "FORCE_NATIVE"
# The projection handling policy for layers that should use the projection
# information from the underlying storage mechanism directly, and ignore the
## projection setting.

REPROJECT = "REPROJECT"
# The projection handling policy for layers that should use the projection
# information from the underlying storage mechanism to reproject to the
# configured projection.


class ResourceInfo:

    def __init__(self):
        self.dom = None
        self.dirty = dict()

    def fetch(self):
        self.dom = self.catalog.get_xml(self.href)

    def clear(self):
        self.dirty = dict()

    def refresh(self):
        self.clear()
        self.fetch()

    def serialize(self, builder):
        # GeoServer will disable the resource if we omit the <enabled> tag,
        # so force it into the dirty dict before writing
        if hasattr(self, "enabled"):
            self.dirty['enabled'] = self.enabled
        if hasattr(self, "advertised"):
            self.dirty['advertised'] = self.advertised
        for k, writer in self.writers.items():
            if k in self.dirty:
                writer(builder, self.dirty[k])

    def message(self):
        builder = TreeBuilder()
        builder.start(self.resource_type, dict())
        self.serialize(builder)
        builder.end(self.resource_type)
        msg = tostring(builder.close())
        return msg


class DimensionInfo:

    _lookup = (
        ('seconds', 1),
        ('minutes', 60),
        ('hours', 3600),
        ('days', 86400),
        ('months', 2628000000),  # this what geoserver computes for 1 month
        ('years', 31536000000)
    )

    def __init__(self, name, enabled, presentation, resolution, units,
                 unit_symbol, strategy=None, attribute=None,
                 end_attribute=None, reference_value=None):
        self.name = name
        self.enabled = enabled
        self.attribute = attribute
        self.end_attribute = end_attribute
        self.presentation = presentation
        self.resolution = resolution
        self.units = units
        self.unit_symbol = unit_symbol
        self.strategy = strategy
        self.referenceValue = reference_value

    def _multiplier(self, name):
        name = name.lower()
        found = [i[1] for i in self._lookup if i[0] == name]
        if not found:
            raise ValueError('invalid multipler: {}'.format(name))
        return found[0]

    def resolution_millis(self):
        """
        if set, get the value of resolution in milliseconds
        """
        if self.resolution is None or not isinstance(self.resolution, str):
                return self.resolution
        val, mult = self.resolution.split(' ')
        return int(float(val) * self._multlipier(mult) * 1000)

    def resolution_str(self):
        """
        if set, get the value of resolution as '<n> <period>s',
        for example: '8 seconds'
        """
        if self.resolution is None or isinstance(self.resolution, str):
            return self.resolution
        seconds = self.resolution / 1000.
        biggest = self._lookup[0]
        for entry in self._lookup:
            if seconds < entry[1]:
                break
            biggest = entry
        val = seconds / biggest[1]
        if val == int(val):
            val = int(val)
        return '{} {}'.format(val, biggest[0])


def xml_property(path, converter=lambda x: x.text, default=None):
    def getter(self):
        try:
            if path in self.dirty:
                return self.dirty[path]
            else:
                if self.dom is None:
                    self.fetch()
                node = self.dom.find(path)
                if node is not None:
                    return converter(self.dom.find(path))
                return default
        except Exception as e:
            raise AttributeError(e)


    def setter(self, value):
        self.dirty[path] = value

    def delete(self):
        self.dirty[path] = None

    return property(getter, setter, delete)


def write_string(name):
    def write(builder, value):
        builder.start(name, dict())
        if value is not None:
            builder.data(value)
        builder.end(name)
    return write


def write_metadata(name):
    def write(builder, metadata):
        builder.start(name, dict())
        for k, v in metadata.items():
            builder.start("entry", dict(key=k))
            if k in ['time', 'elevation'] or k.startswith('custom_dimension'):
                dimension_info(builder, v)
            elif k == 'DynamicDefaultValues':
                dynamic_default_values_info(builder, v)
            elif k == 'JDBC_VIRTUAL_TABLE':
                jdbc_virtual_table(builder, v)
            else:
                builder.data(v)
            builder.end("entry")
        builder.end(name)
    return write


def write_bool(name):
    def write(builder, b):
        builder.start(name, dict())
        builder.data("true" if b and b != "false" else "false")
        builder.end(name)
    return write


def write_bbox(name):
    def write(builder, b):
        builder.start(name, dict())
        bbox_xml(builder, b)
        builder.end(name)
    return write


def write_string_list(name):
    def write(builder, words):
        builder.start(name, dict())
        if words:
            words = [w for w in words if len(w) > 0]
            for w in words:
                builder.start("string", dict())
                builder.data(w)
                builder.end("string")
        builder.end(name)
    return write


def write_dict(name):
    def write(builder, pairs):
        builder.start(name, dict())
        for k, v in pairs.items():
            if k == 'port':
                v = str(v)
            builder.start("entry", dict(key=k))
            builder.data(v)
            builder.end("entry")
        builder.end(name)
    return write


def bbox(node):
    if node is not None:
        minx = node.find("minx")
        maxx = node.find("maxx")
        miny = node.find("miny")
        maxy = node.find("maxy")
        crs = node.find("crs")
        crs = crs.text if crs is not None else None
        if None not in [minx, maxx, miny, maxy]:
            return minx.text, maxx.text, miny.text, maxy.text, crs
        else:
            return None
    else:
        return None


def bbox_xml(builder, box):
    minx, maxx, miny, maxy, crs = box
    builder.start("minx", dict())
    builder.data(minx)
    builder.end("minx")
    builder.start("maxx", dict())
    builder.data(maxx)
    builder.end("maxx")
    builder.start("miny", dict())
    builder.data(miny)
    builder.end("miny")
    builder.start("maxy", dict())
    builder.data(maxy)
    builder.end("maxy")
    if crs is not None:
        builder.start("crs", {"class": "projected"})
        builder.data(crs)
        builder.end("crs")


def md_dynamic_default_values_info(name, node):
    """
    Extract metadata Dynamic Default Values from an xml node
    """
    configurations = node.find("configurations")
    if configurations is not None:
        configurations = []
        for n in node.findall("configuration"):
            dimension = n.find("dimension")
            if dimension is not None:
                dimension = dimension.text
            policy = n.find("policy")
            policy = policy.text if policy is not None else None
            defaultValueExpression = n.find("defaultValueExpression")
            if defaultValueExpression is not None:
                defaultValueExpression = defaultValueExpression.text
            configurations.append(DynamicDefaultValuesConfiguration(
                dimension,
                policy,
                defaultValueExpression
            ))
    return DynamicDefaultValues(name, configurations)


def md_dimension_info(name, node):
    """Extract metadata Dimension Info from an xml node"""
    child_text = lambda child_name: getattr(node.find(child_name),
                                            'text', None)
    resolution = child_text('resolution')
    default_value = node.find("defaultValue")
    if default_value is not None:
        strategy = default_value.find("strategy")
    else:
        strategy = None
    strategy = strategy.text if strategy is not None else None
    return DimensionInfo(
        name,
        child_text('enabled') == 'true',
        child_text('presentation'),
        int(resolution) if resolution else None,
        child_text('units'),
        child_text('unitSymbol'),
        # child_text('strategy'),
        strategy,
        child_text('attribute'),
        child_text('endAttribute'),
        child_text('referenceValue'),
    )


class JDBCVirtualTableGeometry(object):

    def __init__(self, _name, _type, _srid):
        self.name = _name
        self.type = _type
        self.srid = _srid


class JDBCVirtualTableParam(object):

    def __init__(self, _name, _defaultValue, _regexpValidator):
        self.name = _name
        self.defaultValue = _defaultValue
        self.regexpValidator = _regexpValidator


class JDBCVirtualTable(object):
    def __init__(self, _name, _sql, _escapeSql, _geometry,
                 _keyColumn=None, _parameters=None):
        self.name = _name
        self.sql = _sql
        self.escapeSql = _escapeSql
        self.geometry = _geometry
        self.keyColumn = _keyColumn
        self.parameters = _parameters


def md_jdbc_virtual_table(key, node):
    """
    Extract metadata JDBC Virtual Tables from an xml node
    """
    name = node.find("name")
    sql = node.find("sql")
    escapeSql = node.find("escapeSql")
    escapeSql = escapeSql.text if escapeSql is not None else None
    keyColumn = node.find("keyColumn")
    keyColumn = keyColumn.text if keyColumn is not None else None
    n_g = node.find("geometry")
    geometry = JDBCVirtualTableGeometry(
        n_g.find("name"),
        n_g.find("type"),
        n_g.find("srid")
    )
    parameters = []
    for n_p in node.findall("parameter"):
        p_name = n_p.find("name")
        p_defaultValue = n_p.find("defaultValue")
        if p_defaultValue is not None:
            p_defaultValue = p_defaultValue.text
        p_regexpValidator = n_p.find("regexpValidator")
        if p_regexpValidator is not None:
            p_regexpValidator = p_regexpValidator.text
        parameters.append(JDBCVirtualTableParam(
            p_name,
            p_defaultValue,
            p_regexpValidator
        ))
    return JDBCVirtualTable(
        name,
        sql,
        escapeSql,
        geometry,
        keyColumn,
        parameters
    )


def md_entry(node):
    """
    Extract metadata entries from an xml node
    """
    key = None
    value = None
    if 'key' in node.attrib:
        key = node.attrib['key']
    else:
        key = None

    if key in ['time', 'elevation'] or key.startswith('custom_dimension'):
        value = md_dimension_info(key, node.find("dimensionInfo"))
    elif key == 'DynamicDefaultValues':
        value = md_dynamic_default_values_info(key, node.find("DynamicDefaultValues"))
    elif key == 'JDBC_VIRTUAL_TABLE':
        value = md_jdbc_virtual_table(key, node.find("virtualTable"))
    else:
        value = node.text

    if None in [key, value]:
        return None
    else:
        return key, value


class DynamicDefaultValuesConfiguration(object):
    def __init__(self, dimension, policy, default_value_expression):
        self.dimension = dimension
        self.policy = policy
        self.defaultValueExpression = default_value_expression


class DynamicDefaultValues(object):
    def __init__(self, name, configurations):
        self.name = name
        self.configurations = configurations


def md_dynamic_default_values_info(name, node):
    """Extract metadata Dynamic Default Values from an xml node"""
    configurations = node.find("configurations")
    if configurations is not None:
        configurations = []
        for n in node.findall("configuration"):
            dimension = n.find("dimension")
            dimension = dimension.text if dimension is not None else None
            policy = n.find("policy")
            policy = policy.text if policy is not None else None
            default_value_expression = n.find("defaultValueExpression")
            if default_value_expression.text is not None:
                default_value_expression = default_value_expression.text
            else:
                default_value_expression = None
            conf = DynamicDefaultValuesConfiguration(
                dimension,
                policy,
                default_value_expression
            )
            configurations.append(conf)
    return DynamicDefaultValues(name, configurations)


def metadata(node):
    if node is not None:
        return dict(md_entry(n) for n in node.findall("entry"))


def string_list(node):
    if node is not None:
        return [n.text for n in node.findall("string")]


def attribute_list(node):
    if node is not None:
        return [n.text for n in node.findall("attribute/name")]


def key_value_pairs(node):
    if node is not None:
        return dict((entry.attrib['key'], entry.text)
                    for entry in node.findall("entry"))


def dimension_info(builder, metadata):
    if isinstance(metadata, DimensionInfo):
        builder.start("dimensionInfo", dict())
        builder.start("enabled", dict())
        builder.data("true" if metadata.enabled else "false")
        builder.end("enabled")
        if metadata.presentation is not None:
            accepted = ['LIST', 'DISCRETE_INTERVAL', 'CONTINUOUS_INTERVAL']
            if metadata.presentation not in accepted:
                msg = "metadata.presentation must be one of the following {}"
                raise ValueError(msg.format(accepted))
            else:
                builder.start("presentation", dict())
                builder.data(metadata.presentation)
                builder.end("presentation")
        if metadata.attribute is not None:
            builder.start("attribute", dict())
            builder.data(metadata.attribute)
            builder.end("attribute")
        if metadata.end_attribute is not None:
            builder.start("endAttribute", dict())
            builder.data(metadata.end_attribute)
            builder.end("endAttribute")
        if metadata.resolution is not None:
            builder.start("resolution", dict())
            builder.data(str(metadata.resolution_millis()))
            builder.end("resolution")
        if metadata.units is not None:
            builder.start("units", dict())
            builder.data(metadata.units)
            builder.end("units")
        if metadata.unitSymbol is not None:
            builder.start("unitSymbol", dict())
            builder.data(metadata.unitSymbol)
            builder.end("unitSymbol")
        if metadata.strategy is not None:
            builder.start("defaultValue", dict())
            builder.start("strategy", dict())
            builder.data(metadata.strategy)
            builder.end("strategy")
            if metadata.referenceValue:
                builder.start("referenceValue", dict())
                builder.data(metadata.referenceValue)
                builder.end("referenceValue")
            builder.end("defaultValue")
        builder.end("dimensionInfo")


def dynamic_default_values_info(builder, metadata):
    if isinstance(metadata, DynamicDefaultValues):
        builder.start("DynamicDefaultValues", dict())
        if metadata.configurations is not None:
            builder.start("configurations", dict())
            for c in metadata.configurations:
                builder.start("configuration", dict())
                if c.dimension is not None:
                    builder.start("dimension", dict())
                    builder.data(c.dimension)
                    builder.end("dimension")
                if c.policy is not None:
                    builder.start("policy", dict())
                    builder.data(c.policy)
                    builder.end("policy")
                if c.defaultValueExpression is not None:
                    builder.start("defaultValueExpression", dict())
                    builder.data(c.defaultValueExpression)
                    builder.end("defaultValueExpression")
                builder.end("configuration")
            builder.end("configurations")
        builder.end("DynamicDefaultValues")


def jdbc_virtual_table(builder, metadata):
    if isinstance(metadata, JDBCVirtualTable):
        builder.start("virtualTable", dict())
        # name
        builder.start("name", dict())
        builder.data(metadata.name)
        builder.end("name")
        # sql
        builder.start("sql", dict())
        builder.data(metadata.sql)
        builder.end("sql")
        # escapeSql
        builder.start("escapeSql", dict())
        builder.data(metadata.escapeSql)
        builder.end("escapeSql")
        # keyColumn
        if metadata.keyColumn is not None:
            builder.start("keyColumn", dict())
            builder.data(metadata.keyColumn)
            builder.end("keyColumn")
        # geometry
        if metadata.geometry is not None:
            g = metadata.geometry
            builder.start("geometry", dict())
            if g.name is not None:
                builder.start("name", dict())
                builder.data(g.name)
                builder.end("name")
            if g.type is not None:
                builder.start("type", dict())
                builder.data(g.type)
                builder.end("type")
            if g.srid is not None:
                builder.start("srid", dict())
                builder.data(g.srid)
                builder.end("srid")
            builder.end("geometry")
        # parameters
        if metadata.parameters is not None:
            for p in metadata.parameters:
                builder.start("parameter", dict())
                if p.name is not None:
                    builder.start("name", dict())
                    builder.data(p.name)
                    builder.end("name")
                if p.defaultValue is not None:
                    builder.start("defaultValue", dict())
                    builder.data(p.defaultValue)
                    builder.end("defaultValue")
                if p.regexpValidator is not None:
                    builder.start("regexpValidator", dict())
                    builder.data(p.regexpValidator)
                    builder.end("regexpValidator")
                builder.end("parameter")
        builder.end("virtualTable")


def prepare_upload_bundle(name, data):
    """
    GeoServer's REST API uses ZIP archives as containers for file formats such
    as Shapefile and WorldImage which include several 'boxcar' files alongside
    the main data.  In such archives, GeoServer assumes that all of the relevant
    files will have the same base name and appropriate extensions, and live in
    the root of the ZIP archive.  This method produces a zip file that matches
    these expectations, based on a basename, and a dict of extensions to paths or
    file-like objects. The client code is responsible for deleting the zip
    archive when it's done.
    """
    fd, path = mkstemp()
    zip_file = ZipFile(path, 'w')
    for ext, stream in data.items():
        fname = "{}.{}".format(name, ext)
        if isinstance(stream, str):
            zip_file.write(stream, fname)
        else:
            zip_file.writestr(fname, stream.read())
    zip_file.close()
    os.close(fd)
    return path


def _decode_list(data):
    rv = []
    for item in data:
        if isinstance(item, str):
            item = item.encode('utf-8')
        elif isinstance(item, list):
            item = _decode_list(item)
        elif isinstance(item, dict):
            item = _decode_dict(item)
        rv.append(item)
    return rv


def _decode_dict(data):
    rv = {}
    for key, value in data.items():
        if isinstance(key, str):
            key = key.encode('utf-8')
        if isinstance(value, str):
            value = value.encode('utf-8')
        elif isinstance(value, list):
            value = _decode_list(value)
        elif isinstance(value, dict):
            value = _decode_dict(value)
        rv[key] = value
    return rv



# coding: utf-8

from urllib.parse import urljoin

from geoserver.support import ResourceInfo, write_string, write_bbox, \
    xml_property, bbox
from geoserver import settings


def _maybe_text(n):
    if n is None:
        return None
    else:
        return n.text


def _layer_list(node, element):
    if node is not None:
        return [_maybe_text(n.find("name")) for n in node.findall(element)]


def _style_list(node):
    if node is not None:
        return [_maybe_text(n.find("name")) for n in node.findall("style")]


def _write_layers(builder, layers, parent, element, attributes):
    builder.start(parent, dict())
    for l in layers:
        builder.start(element, attributes or dict())
        if l is not None:
            builder.start("name", dict())
            builder.data(l)
            builder.end("name")
        builder.end(element)
    builder.end(parent)


def _write_styles(builder, styles):
    builder.start("styles", dict())
    for s in styles:
        builder.start("style", dict())
        if s is not None:
            builder.start("name", dict())
            builder.data(s)
            builder.end("name")
        builder.end("style")
    builder.end("styles")


class LayerGroup(ResourceInfo):
    """    Represents a layer group in geoserver
    """

    resource_type = "layerGroup"
    save_method = settings.PUT

    def __init__(self, catalog, name, workspace=None):
        super(LayerGroup, self).__init__()
        assert isinstance(name, str)
        self.catalog = catalog
        self.name = name
        self.workspace = workspace
        # the XML format changed in 2.3.x - the element listing all the layers
        # and the entries themselves have changed
        if self.catalog.gsversion() == "2.2.x":
            parent, element, attributes = "layers", "layer", None
        else:
            parent = "publishables"
            element = "published"
            attributes = {'type': 'layer'}
        self._layer_parent = parent
        self._layer_element = element
        self._layer_attributes = attributes
        self.writers = {
            'name': write_string("name"),
            'styles': _write_styles,
            'layers': lambda b, l: _write_layers(b, l, parent,
                                                 element, attributes),
            'bounds': write_bbox("bounds"),
            'workspace': write_string("workspace")
        }

    @property
    def href(self):
        path_parts = "layergroups/{}.xml".format(self.name)
        if self.workspace is not None:
            workspace_name = getattr(self.workspace, 'name', self.workspace)
            path_parts = "workspaces/{}/{}".format(workspace_name, path_parts)
        return urljoin(
            self.catalog.service_url,
            path_parts
        )
        return url(self.catalog.service_url, path_parts)

    styles = xml_property("styles", _style_list)
    bounds = xml_property("bounds", bbox)

    @property
    def layers(self):
        if "layers" in self.dirty:
            return self.dirty["layers"]
        else:
            if self.dom is None:
                self.fetch()
            node = self.dom.find(self._layer_parent)
            if node is not None:
                return _layer_list(node, self._layer_element)
            return None

    @layers.setter
    def _layers_setter(self, value):
        self.dirty["layers"] = value

    @layers.deleter
    def _layers_delete(self):
        self.dirty["layers"] = None

    def __str__(self):
        return "<LayerGroup {}>".format(self.name)

    __repr__ = __str__


class UnsavedLayerGroup(LayerGroup):

    save_method = settings.POST

    def __init__(self, catalog, name, layers, styles, bounds, workspace=None):
        super(UnsavedLayerGroup, self).__init__(
            catalog,
            name,
            workspace=workspace
        )
        if bounds is None:
            bounds = ("-180", "180", "-90", "90", "EPSG:4326")
        self.dirty.update(name=name, layers=layers, styles=styles,
                          bounds=bounds, workspace=workspace)

    @property
    def href(self):
        path_parts = 'layergroups'
        if self.workspace is not None:
            workspace_name = getattr(self.workspace, 'name', self.workspace)
            path_parts = "workspaces/{}/{}".format(workspace_name, path_parts)
        return urljoin(
            self.catalog.service_url,
            "{}?name={}".format(path_parts, self.name)
        )
